/*
    Avatar provider - based on Microblog dataengine implementation
    Copyright (C) 2008  Aaron Seigo <aseigo@kde.org>
    Copyright (C) 2012  Martin Klapetek <martin.klapetek@gmail.com>

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
    License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to the
    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
*/


#include "avatarprovider.h"

#include <KDebug>
#include <KIO/Job>
#include <KImageCache>

#include <QPainter>

class Akonadi::AvatarProviderPrivate {
public:

    struct QueuedJobHelper {
        QString who;
        KUrl url;
        bool polishImage;
    };

    /**
     * Makes the corners of the image rounded
     */
    QImage polishImage( const QImage &img )
    {
        QImage roundedImage = QImage( QSize( 48*4, 48*4 ), QImage::Format_ARGB32_Premultiplied );
        roundedImage.fill( Qt::transparent );
        QPainter p;
        p.begin( &roundedImage );
        QPainterPath clippingPath;
        QRectF imgRect = QRectF( QPoint( 0,0 ), roundedImage.size() );
        clippingPath.addRoundedRect( imgRect, 24, 24 );
        p.setClipPath( clippingPath );
        p.setClipping( true );
        p.drawImage( QRectF( QPointF( 0, 0 ), roundedImage.size() ), img );
        return roundedImage;
    }

    ///All active jobs
    QHash<KJob *, QString> jobs;

    ///Received data from the jobs
    QHash<KJob *, QByteArray> jobData;

    ///Number of running jobs
    int runningJobs;

    ///Queued jobs when the limit is reached
    QList<QueuedJobHelper> queuedJobs;

    ///String list containing the list of loaded persons avatars
    QStringList pendingPersons;

    ///KImageCache instance
    KImageCache *imageCache;
};

Akonadi::AvatarProvider::AvatarProvider( QObject *parent )
    : QObject( parent ),
      d_ptr(new AvatarProviderPrivate)
{
    Q_D(AvatarProvider);
    d->imageCache = 0;
    d->runningJobs = 0;
}

Akonadi::AvatarProvider::~AvatarProvider()
{
    Q_D(AvatarProvider);
    delete d;
}

void Akonadi::AvatarProvider::loadAvatar( const QString &who, const KUrl &url, bool polishImage )
{
    Q_D(AvatarProvider);

    if ( who.isEmpty() ) {
        return;
    }
    if ( !d->imageCache ) {
        d->imageCache = new KImageCache( "plasma_engine_preview", 10485760 ); // Re-use previewengine's cache
    }

    const QString cacheKey = who + "@" + url.pathOrUrl();

    // Make sure we only start one job per user
    if ( d->pendingPersons.contains( cacheKey ) ) {
        kDebug() << "Job for" << who << "already running, returning";
        return;
    }

    // Check if the image is in the cache, if so emit it and return
    QImage preview = QImage( QSize( 48, 48 ), QImage::Format_ARGB32_Premultiplied ); //FIXME: handle the size better
    preview.fill( Qt::transparent );

    if ( d->imageCache->findImage( cacheKey, &preview ) ) {
        // cache hit
        kDebug() << "Image for" << who << "already in cache, emitting result";
        Q_EMIT imageLoaded( who, url, polishImage ? d->polishImage( preview ) : preview );
        return;
    }

    if ( !url.isValid() ) {
        kDebug() << "Invalid url, returning";
        return;
    }

    kDebug() << "No cache, fetching image for" << who;

    d->pendingPersons << cacheKey;
    //FIXME: since kio_http bombs the system with too many request put a temporary
    // arbitrary limit here, revert as soon as BUG 192625 is fixed
    // Note: seems fixed.
    if ( d->runningJobs < 500) {
        d->runningJobs++;
        kDebug( ) << "Starting fetch job for" << who;
        KIO::Job *job = KIO::get( url, KIO::NoReload, KIO::HideProgressInfo );
        job->setAutoDelete( true );
        d->jobs[job] = who;
        connect( job, SIGNAL( data( KIO::Job*,QByteArray ) ),
                 this, SLOT( recv( KIO::Job*,QByteArray ) ) );
        connect( job, SIGNAL( result( KJob* ) ),
                 this, SLOT( result( KJob* ) ) );
        //the url needs to be stored explicitly because for example facebook redirects randomly between its servers
        //causing the url to be different everytime, which makes storing the job's url in cache impossible
        //therefore we set it here and use this when saving to the cache instead of job->url()
        job->setProperty( "imageUrl", url.pathOrUrl() );
        job->setProperty( "polishImage", polishImage );
        job->start();
    } else {
        kDebug() << "Queuing job for" << who;
        AvatarProviderPrivate::QueuedJobHelper helper;
        helper.who = who;
        helper.url = url;
        helper.polishImage = polishImage;
        d->queuedJobs.append( helper );
    }
}

void Akonadi::AvatarProvider::recv( KIO::Job *job, const QByteArray &data )
{
    Q_D(AvatarProvider);
    d->jobData[job] += data;
}

void Akonadi::AvatarProvider::result( KJob *job )
{
    Q_D(AvatarProvider);
    if ( !d->jobs.contains( job ) ) {
        kDebug() << "Tried to handle unknown job, returning...";
        return;
    }

    d->runningJobs--;

    if ( d->queuedJobs.count() > 0 ) {
        AvatarProviderPrivate::QueuedJobHelper helper = d->queuedJobs.takeFirst();
        loadAvatar( helper.who, helper.url, helper.polishImage );
    }

    if ( job->error()) {
        // TODO: error handling
        KIO::TransferJob* kiojob = dynamic_cast<KIO::TransferJob*>( job );
        kError() << "Image job for" << d->jobs.value( job ) << "returned error:" << kiojob->errorString();
    } else {
        const QString who = d->jobs.value( job );

        QImage image;
        image.loadFromData( d->jobData.value( job ) );
        KIO::TransferJob* kiojob = dynamic_cast<KIO::TransferJob*>( job );
        const QString cacheKey = who + "@" + kiojob->property( "imageUrl" ).toString();

        kDebug() << "Downloaded image for" << who << "(key:" << cacheKey << ")";

        d->imageCache->insertImage( cacheKey, image );
        d->pendingPersons.removeAll( cacheKey );

        bool polishImage = job->property( "polishImage" ).toBool();

        Q_EMIT imageLoaded( who, kiojob->property( "imageUrl" ).toUrl(), polishImage ? d->polishImage( image ) : image );
    }

    d->jobs.remove( job );
    d->jobData.remove( job );
}

void Akonadi::AvatarProvider::abortAllJobs()
{
    Q_D(AvatarProvider);
    Q_FOREACH( KJob *job, d->jobs.keys() ) {
        job->kill();
        d->jobs.remove( job );
        d->jobData.remove( job );
    }
}
